/*   implementation of the own backend 
 *  to use UdmSearch without third-party libraries 
 */

#include "udm_config.h"


#ifdef HAVE_FILES
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_db.h"
#include "udm_charset.h"
#include "udm_spell.h"
#include "udm_robots.h"
#include "udm_mutex.h"
#include "udm_utils.h"
#include "udm_log.h"

char DBVersion[] = "built-in";
char *DBHost=NULL;
char *DBName=NULL;
char *DBUser=NULL;
char *DBPass=NULL;
int  DBPort=0;
int  DBType=0;
int  DBMode=0;
int  DBUseLock=0;

static char * ul = NULL;

char * escstr(char *x,char *y){
	return(strcpy(x,y));
}


/************************* Backen functions *****************/

typedef struct txt_db {
	int connected;
	FILE * dict;
	FILE * url;
	FILE * robots;
	FILE * stopwords;
	int  open_mode;
	int  errcode;
	char errmsg[1024];
} DB;

static void CloseDB(void * vdb){
DB * db;
	db=(DB*)vdb;
	if(db->dict)fclose(db->dict);
	if(db->url)fclose(db->url);
	if(db->robots)fclose(db->robots);
	if(db->stopwords)fclose(db->stopwords);
}

void UdmFreeDB(void * vdb){
DB * db;
	db=(DB*)vdb;
	if(db->connected)CloseDB(vdb);
	free(vdb);
}

static int InitDB(DB *db){
char * dictfile = UDM_VAR_DIR "/dict.txt";
char * urlfile = UDM_VAR_DIR "/url.txt";
char * fmode;

	if(db->connected)
		return(0);

	switch(db->open_mode){
		case UDM_OPEN_MODE_READ:
			fmode="r";
			break;
		case UDM_OPEN_MODE_WRITE:
		default:
			fmode="w";
	}

	db->dict=fopen(dictfile,fmode);
	db->connected=1;
	if(!db->dict){
		sprintf(db->errmsg,"Can't open dict file '%s' (%s)",dictfile, strerror(errno));
		db->errcode=1;
		return(1);
	}
	db->url=fopen(urlfile,fmode);
	if(!db->url){
		sprintf(db->errmsg,"Can't open URL file '%s' (%s)", urlfile, strerror(errno));
		db->errcode=1;
		return(1);
	}
	return(0);
}

void * UdmAllocDB(int mode){
DB * db;
	db=(DB*)malloc(sizeof(DB));
	db->dict=NULL;
	db->url=NULL;
	db->robots=NULL;
	db->stopwords=NULL;
	db->errcode=0;
	db->errmsg[0]=0;
	db->open_mode=mode;
	db->connected=0;
	return((void*)db);
}
char * UdmDBErrorMsg(void *db){
	return(((DB*)db)->errmsg);
}
int UdmDBErrorCode(void *db){
	return(((DB*)db)->errcode);
}


/*************  Words ****************************/
int UdmDeleteWordFromURL(UDM_INDEXER* Indexer,int url_id){
	return(0);
}
int UdmStoreWords(UDM_INDEXER * Indexer, int url_id){
int i;
FILE *f;
	if (InitDB(((DB*)(Indexer->db)))){
	    UdmLog(Indexer->handle, UDM_LOG_ERROR, "Error: %s", ((DB*)(Indexer->db))->errmsg);
	    exit(1);
	}
	f=((DB*)(Indexer->db))->dict;
	for(i=0;i<Indexer->nwords;i++){
		if(Indexer->Word[i].count){
			fprintf(f,"%d\t%d\t%s\n",url_id,
				Indexer->Word[i].count,
				Indexer->Word[i].word);
		}
	}
	return(IND_OK);
}

/************** stopwords **********************/
int load_stop_list(void * db){
	return(0);
}
UDM_STOPWORD * UdmIsStopWord(char *word){
	return(NULL);
}



/************ URLs stuff *******************************/

static int n_urls=0;
static UDM_DOCUMENT* URLs=NULL;


UDM_DOCUMENT * UdmGetDocInfo(UDM_INDEXER* Indexer,int expired_first){
int i;

UDM_DOCUMENT * Result;

	for(i=0;i<n_urls;i++){
		if(!URLs[i].indexed){
			URLs[i].indexed=1;
			Result=(UDM_DOCUMENT *)malloc(sizeof(UDM_DOCUMENT));
			Result->content_type=NULL;
			Result->title=NULL;
			Result->text=NULL;
			Result->last_index_time=NULL;
			Result->next_index_time=NULL;
			Result->keywords=NULL;
			Result->description=NULL;
			Result->crc=NULL;

			Result->url=strdup(URLs[i].url);
			Result->url_id=URLs[i].url_id;
			Result->size=0;
			Result->status=0;
			Result->last_index_time=strdup("0");
			Result->hops=URLs[i].hops;
			Result->crc=strdup("___");
			Result->last_modified=strdup("");
			return(Result);
		}
	}
	return(NULL);
}
int UdmAddURL(UDM_INDEXER* Indexer,char *url,int referrer,int hops, char * msg_id){
	if(!n_urls){
		URLs=(UDM_DOCUMENT *)malloc(sizeof(UDM_DOCUMENT));
	}else{
		URLs=(UDM_DOCUMENT *)realloc(URLs,sizeof(UDM_DOCUMENT)*(n_urls+1));
	}
	URLs[n_urls].url_id=n_urls+1;
	URLs[n_urls].indexed=0;
	URLs[n_urls].hops=hops;
	URLs[n_urls].referrer=referrer;
	URLs[n_urls].url=strdup(url);
	n_urls++;
	return(IND_OK);
}
int UdmDeleteUrl(UDM_INDEXER* Indexer,int url_id){
	return(IND_OK);
}
int UdmUpdateUrl(UDM_INDEXER* Indexer,int url_id,int status,int period){
	return(IND_OK);
}

#define XVAL(x)		(!x?"-":!x[0]?"-":x)
int UdmLongUpdateUrl(UDM_INDEXER* Indexer,
	int url_id,int status,int is_changed,
	int size,int period,int hint,int index,
	char *last_modified_escaped,
	char *text_escaped,
	char *title_escaped,
	char *content_type,
	char *keywords_escaped,
	char *descript_escaped,
	char *digest,
	char *lang
#ifdef NEWS_EXT
	,
	char *hd_date,
	char *hd_subj,
	char *hd_from,
	char *hd_group,
	char *hd_ref,
	char *msg_id
#endif
	){
	
	int i;FILE *f;

	if (InitDB(((DB*)(Indexer->db)))){
	    UdmLog(Indexer->handle, UDM_LOG_ERROR, "Error: %s", ((DB*)(Indexer->db))->errmsg);
	    exit(1);
	}

	f=((DB*)(Indexer->db))->url;

	for(i=0;i<n_urls;i++)
	if(URLs[i].url_id==url_id){

		fprintf(f,"%d\t%d\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
			url_id,hint,status,size,
			XVAL(content_type),
			XVAL(lang),
			XVAL(digest),
			XVAL(URLs[i].url),
			XVAL(last_modified_escaped),
			XVAL(text_escaped),
			XVAL(title_escaped),
			XVAL(keywords_escaped),
			XVAL(descript_escaped));
		break;
	}
	return(IND_OK);
}



/************* Clones ***************************/
int UdmFindOrigin(UDM_INDEXER* Indexer,char *crc){
	return(0);
}
int UdmUpdateClone(UDM_INDEXER* Indexer,int url_id,int status,int period,
	char *content_type,char *last_modified_escaped,char *digest){
	return(IND_OK);
}
UDM_DOCUMENT * UdmCloneList(void * db,char *crc){
	return(NULL);
}
int UdmRegisterChild(UDM_INDEXER *Indexer, int parent_id, int child_id){
	return 0;
}
int UdmFindMsgID(UDM_INDEXER *Indexer, const char * msg_id){
	return 0;
}



/************* Clearing **************************/
int UdmDeleteAllFromDict(UDM_INDEXER* Indexer){
	return(IND_OK);
}
int UdmDeleteAllFromUrl(UDM_INDEXER* Indexer){
	return(IND_OK);
}
int UdmClearDB(UDM_INDEXER* Indexer){
	return(IND_OK);
}



/************** Robots.txt stuff *****************/
int UdmLoadRobots(UDM_INDEXER *Indexer){
	return(IND_OK);
}
int UdmDeleteRobotsFromHost(UDM_INDEXER* Indexer,char *hostinfo){
	return(IND_OK);
}
int UdmAddRobotsToHost(UDM_INDEXER* Indexer,char *hostinfo,char *s){
int nrobots;
	if(LockProc)LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
	if(!Robots){
		nrobots=0;
		Robots=(UDM_ROBOT *)malloc((2)*sizeof(UDM_ROBOT));
	}else{
		for(nrobots=0;Robots[nrobots].hostinfo;nrobots++);
		Robots=(UDM_ROBOT *)realloc(Robots,(nrobots+2)*sizeof(UDM_ROBOT));
	}
	Robots[nrobots].hostinfo=strdup(hostinfo);
	Robots[nrobots].path=strdup(s);
	nrobots++;
	Robots[nrobots].hostinfo=Robots[nrobots].path=NULL;

	if(LockProc)LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
	return(IND_OK);
}


typedef struct wrd{
	int  url_id;
	int  count;
	int  weight;
} SEARCHWORD;

static int cmpword(const void *s1,const void *s2){
int res;
	if(!(res=(((SEARCHWORD*)s2)->count)-(((SEARCHWORD*)s1)->count)))
		if(!(res=(((SEARCHWORD*)s2)->weight-(((SEARCHWORD*)s1)->weight))))
			if(!(res=(((SEARCHWORD*)s1)->url_id-(((SEARCHWORD*)s2)->url_id))));
	return(res);
}
static int cmpurlid(const void *s1,const void *s2){
	return(((SEARCHWORD*)s2)->url_id-((SEARCHWORD*)s1)->url_id);
}


/**************** search stuff *******************/
UDM_DOCUMENT * UdmFind(void * db,char *q,int np,int ps, int search_mode,int sort_order,char *wordinfo,int *found){
char *w,*lasttok;
char *words[100];
char str[UDMSTRSIZ];
char curword[100];
int weight;
int url_id;
int num_words=0;
int i,j;
SEARCHWORD * wrd=NULL;
int nwrd=0;
UDM_DOCUMENT * r;
int first;
int docnum=0;
int skip=0;

	if(InitDB(db))
		return(NULL);
	load_stop_list(db);
	*wordinfo=0;
	if(UdmDBErrorCode(db))return(NULL);
	w=UdmGetWord(q,&lasttok,UdmGetDefaultCharset());
	while(w){
		char ** ww;
		char *rw;

		ww=UdmNormalizeWord(w);
		rw=ww?*ww:w;

		if(UdmIsStopWord(rw)){
			sprintf(UDM_STREND(wordinfo)," %s :stopword",rw);
			w=UdmGetWord(NULL,&lasttok,UdmGetDefaultCharset());
			continue;
		}
		if(num_words<100){
			words[num_words]=strdup(rw);
			num_words++;
		}
		w=UdmGetWord(NULL,&lasttok,UdmGetDefaultCharset());
	}

	while(fgets(str,sizeof(str),((DB*)db)->dict)){

		w=UdmGetToken(str,"\t",&lasttok);
		if(!w)continue;url_id=atoi(w);
		w=UdmGetToken(NULL,"\t",&lasttok);
		if(!w)continue;weight=atoi(w);
		w=UdmGetToken(NULL,"\t\n",&lasttok);
		if(!w)continue;strcpy(curword,w);

		for(i=0;i<num_words;i++){
			if(!strcmp(curword,words[i])){
				if(!nwrd){
					nwrd=1;
					wrd=(SEARCHWORD*)malloc(nwrd*sizeof(SEARCHWORD));
				}else{
					nwrd++;
					wrd=(SEARCHWORD*)realloc(wrd,nwrd*sizeof(SEARCHWORD));
				}
				wrd[nwrd-1].url_id=url_id;
				wrd[nwrd-1].count=1;
				wrd[nwrd-1].weight=weight;
			}
		}
	}

	if(!nwrd)return(0);
	qsort((void*)wrd,nwrd,sizeof(SEARCHWORD),cmpurlid);
	
	j=0;
	for(i=1;i<nwrd;i++){
		if(wrd[j].url_id==wrd[i].url_id){
			wrd[j].weight+=wrd[i].weight;
			wrd[j].count++;
		}else{
			j++;
			wrd[j].url_id=wrd[i].url_id;
			wrd[j].count=wrd[i].count;
			wrd[j].weight=wrd[i].weight;
		}
	}
	nwrd=j+1;

	qsort((void*)wrd,nwrd,sizeof(SEARCHWORD),cmpword);
	if(search_mode==UDM_MOD_ALL){
		for(i=0;i<nwrd;i++)
			if(wrd[i].count<num_words){
				nwrd=i;
				break;
			}
	}
	if(!nwrd)return(0);

	/* sprintf(UDM_STREND(wordinfo)," %s : %d",rw,num); */

	first=np*ps;

	r=(UDM_DOCUMENT*)malloc(sizeof(UDM_DOCUMENT)*(nwrd+1));
	
	while(fgets(str,sizeof(str),((DB*)db)->url)){
		url_id=atoi(str);
		i=0;
		while((i+first)<nwrd && (docnum<ps)){
			r[docnum].url_id=wrd[i+first].url_id;
			r[docnum].rating=wrd[i+first].count;
			r[docnum].content_type=NULL;
			r[docnum].crc=NULL;
			r[docnum].url=NULL;
			r[docnum].last_modified=NULL;
			r[docnum].text=NULL;
			r[docnum].title=NULL;
			r[docnum].keywords=NULL;
			r[docnum].description=NULL;
			if(url_id==wrd[i+first].url_id){
				w=UdmGetToken(str,"\t",&lasttok);
				j=0;
				while(w){
					switch(j){
						case  3: r[docnum].size=atoi(w);break;
						case  4: r[docnum].content_type=strdup(w);break;
						/*case  5: r[docnum].lang=strdup(w);break;*/
						case  6: r[docnum].crc=strdup(w);break;
						case  7: r[docnum].url=strdup(w);break;
						case  8: r[docnum].last_modified=strdup(w);break;
						case  9: r[docnum].text=strdup(w);break;
						case 10: r[docnum].title=strdup(w);break;
						case 11: r[docnum].keywords=strdup(w);break;
						case 12: r[docnum].description=strdup(w);break;	
					}
					w=UdmGetToken(NULL,"\t",&lasttok);
					j++;
				}
				if(ul==NULL){
					docnum++;
				}else{
					if (strstr(r[docnum].url,ul)){
						docnum++;
					}else{
						UDM_FREE(r[docnum].content_type);
						UDM_FREE(r[docnum].crc);
						UDM_FREE(r[docnum].url);
						UDM_FREE(r[docnum].last_modified);
						UDM_FREE(r[docnum].text);
						UDM_FREE(r[docnum].title);
						UDM_FREE(r[docnum].keywords);
						UDM_FREE(r[docnum].description);
						skip++;
					}
				}
			}
			i++;
		}
	}
	r[docnum].url_id=0;
	if(!docnum)return(0);
	*found=nwrd-skip;
	return(r);
}



/***************** Misc functions ***************/
__INDLIB__ int  UdmGetStatistics(){
	return(IND_OK);
}
int UdmMarkForReindex(UDM_INDEXER* Indexer){
	return(IND_OK);
}
__INDLIB__ int UdmAddTagLimit(int tag){
	return(0);
}
__INDLIB__ int UdmAddStatusLimit(int status){
	return(0);
}
__INDLIB__ int UdmAddURLLimit(char *URL){
	UDM_FREE(ul);
	ul=strdup(URL);
	return(0);
}
__INDLIB__ int UdmClearURLLimit(){
	return(0);
}
__INDLIB__ int UdmClearLimits(){
	return(0);
}
__INDLIB__ int UdmGetReferers(UDM_INDEXER* Indexer){
	return(IND_OK);
}

#endif
